package com.yerihyo.yeritools.collections;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.tree.TreeNode;

import com.yerihyo.yeritools.RTTIToolkit;
import com.yerihyo.yeritools.debug.YeriDebug;
import com.yerihyo.yeritools.math.StatisticsToolkit;
import com.yerihyo.yeritools.math.StatisticsToolkit.PredicateQuantifier;
import com.yerihyo.yeritools.text.StringToolkit;
import com.yerihyo.yeritools.text.StringToolkit.StringComparator;

public class CollectionsToolkit {

	public static class CompoundIterator<T> implements Iterator<T> {
		private List<Iterator<? extends T>> iteratorList;
		private int currentIndex = 0;
		
		private CompoundIterator(Iterator<? extends T>[] iteratorArray){
			iteratorList = new ArrayList<Iterator<? extends T>>();
			CollectionsToolkit.addAll(iteratorList, iteratorArray);
		}
		
		private CompoundIterator(Collection<? extends Iterator<? extends T>> iteratorCollection){
			this.iteratorList = new ArrayList<Iterator<? extends T>>(iteratorCollection);
		}
		
		public boolean hasNext(){
			while(!iteratorList.get(currentIndex).hasNext()){
				currentIndex++;
				if(currentIndex>=iteratorList.size()){ return false; }
			}
			return true;
		}
		public T next(){
			hasNext();
			return iteratorList.get(currentIndex).next();
		}
		public void remove(){
			iteratorList.get(currentIndex).remove();
		}
		
		public static <T> Iterator<T> create(Iterator<? extends T>[] iteratorArray){
			return new CompoundIterator<T>(iteratorArray);
		}
		
		public static <T> Iterator<T> create(Collection<? extends Collection<? extends T>> cc){
			List<Iterator<? extends T>> iteratorList = new ArrayList<Iterator<? extends T>>();
			for(Collection<? extends T> c : cc){
				iteratorList.add(c.iterator());
			}
			return new CompoundIterator<T>(iteratorList);
		}
		
		public static <T> Iterator<T> create(Collection<? extends T>[] cArray){
			List<Iterator<? extends T>> iteratorList = new ArrayList<Iterator<? extends T>>();
			for(Collection<? extends T> c : cArray){
				iteratorList.add(c.iterator());
			}
			return new CompoundIterator<T>(iteratorList);
		}
		
	}
	
	
	public static class FilteredIterator<T> implements Iterator<T> {
		private Iterator<T> iterator;
		private Filter<T> filter;
		
		private FilteredIterator(Iterator<T> iterator, Filter<T> filter){
			this.iterator = iterator;
			this.filter = filter;
		}
		
		private transient T cache = null;
		public boolean hasNext(){
			if(cache!=null){ return true; }

			T t = null;
			do{
				if(!iterator.hasNext()){ return false; }
				t = iterator.next();
			}while(!filter.accept(t));
			cache = t;
			return true;
		}
		public T next(){
			hasNext();
			T next = cache;
			cache = null;
			return next;
		}
		public void remove(){
			throw new UnsupportedOperationException();
		}
		
		public static <T> Iterator<T> create(Iterator<T> iterator, Filter<T> filter){
			return new FilteredIterator<T>(iterator, filter);
		}
	}
	
	public static boolean[][] copyOf(boolean[][] tArray){
		return copyOf(tArray, tArray.length);
	}
	public static boolean[][] copyOf(boolean[][] tArray, int length){
		boolean[][] returnValue = Arrays.copyOf(tArray, length);
		for(int i=0; i<length; i++){
			boolean[] row = tArray[i];
			boolean[] newRow = Arrays.copyOf(row, row.length);
			returnValue[i] = newRow;
		}
		return returnValue;
	}
	public static double[][] copyOf(double[][] tArray){
		return copyOf(tArray, tArray.length);
	}
	public static double[][] copyOf(double[][] tArray, int length){
		double[][] returnValue = Arrays.copyOf(tArray, length);
		for(int i=0; i<length; i++){
			double[] row = tArray[i];
			double[] newRow = Arrays.copyOf(row, row.length);
			returnValue[i] = newRow;
		}
		return returnValue;
	}
	
	public static <T> T[][] copyOf(T[][] tArray){
		return copyOf(tArray, tArray.length);
	}
	public static <T> T[][] copyOf(T[][] tArray, int length){
		T[][] returnValue = Arrays.copyOf(tArray, length);
		for(int i=0; i<length; i++){
			T[] row = tArray[i];
			T[] newRow = Arrays.copyOf(row, row.length);
			returnValue[i] = newRow;
		}
		return returnValue;
	}
	
	public static <T> List<T> iteratorToList(Iterator<? extends T> iterator){
		List<T> list = new ArrayList<T>();
		while(iterator.hasNext()){
			list.add(iterator.next());
		}
		return list;
	}
	
	public static interface Filter<T>{
		boolean accept(T t);
	}
	
	public static interface RuleBasedClassifier<T>{
		int classify(T t);
		int getClassCount();
	}
	
	public static class FilterSet<T> extends HashSet<Filter<? super T>>{
		private static final long serialVersionUID = 1L;
		
		public boolean isValidToFilterSet(T t, PredicateQuantifier quantifier){
			boolean stopSearchingValue = quantifier.getStopSearchingValue();
			for(Filter<? super T> filter : this){
				if(filter.accept(t) == stopSearchingValue){ return stopSearchingValue; }
			}
			return !stopSearchingValue;
		}
		 
	}
	public static boolean overlap(int[] range1, int[] range2) {
		YeriDebug.ASSERT(range1.length==2);
		YeriDebug.ASSERT(range2.length==2);
		
		return (range1[1]-range2[0])*(range2[1]-range1[0])>=0;
	}

	public static boolean include(int value, int[] valueRange) {
		YeriDebug.ASSERT(valueRange.length==2);
		return valueRange[0]<=value && valueRange[1]>=value;
	}
	public static <T extends Comparable<? super T>> Comparator<T> createReverseComparator(){
		return new Comparator<T>(){
			public int compare(T t1, T t2){ return -t1.compareTo(t2); }
		};
	}
	
	
	private static class ReverseComparator<T> implements Comparator<T>, Serializable, Cloneable{
		private static final long serialVersionUID = 1L;
		
		private Comparator<T> comparator;
		public ReverseComparator(Comparator<T> comparator){
			this.comparator = comparator;
		}
		
		@Override
		public int compare(T o1, T o2) {
			return -comparator.compare(o1, o2);
		}
	}
	
	public static <T> Comparator<T> createReverseComparator(Comparator<T> comparator){
		return new ReverseComparator<T>(comparator);
	}
	
	public static <T> List<T> filter(Collection<? extends T> collection, Filter<T> filter){
		List<T> returnValue = new ArrayList<T>();
		for(T t : collection){
			if(!filter.accept(t)){
				continue;
			}
			returnValue.add(t);
		}
		return returnValue;
	}
	
	@SuppressWarnings("unchecked")
	public static void getLeafList(List<TreeNode> leafList, TreeNode node){
		if(node.isLeaf()){
			leafList.add(node);
		}else{
			Enumeration<TreeNode> children = node.children();
			while(children.hasMoreElements()){
				TreeNode treeNode = children.nextElement();
				getLeafList(leafList, treeNode);
			}
		}
	}

	public static <K, V> CharSequence mapToString(Map<K, V> map) {
		StringBuffer buffer = new StringBuffer();
		String newLine = StringToolkit.newLine();
		for (Map.Entry<K, V> entry : map.entrySet()) {
			K key = entry.getKey();
			V value = entry.getValue();
			buffer.append(key).append(" = ").append(value).append(newLine);
		}
		return buffer;
	}

	public static <K, V> SortedMap<K, V> subMap(SortedMap<K, V> sortedMap,
			K fromKey, K toKey, boolean includeFrom, boolean includeTo) {
		SortedMap<K, V> returnValue = new TreeMap<K, V>(sortedMap.comparator());

		Iterator<K> iterator = sortedMap.keySet().iterator();

		boolean hasStarted = false;
		while (iterator.hasNext()) {
			K k = iterator.next();
			V v = sortedMap.get(k);
			
			if (k.equals(toKey)) {
//				System.out.println(k);
				if(hasStarted && includeTo){
					returnValue.put(k, v);
				}
				break;
			}
			
			if (!hasStarted && k.equals(fromKey)) {
				hasStarted = true;
				if(includeFrom){
					returnValue.put(k, v);
				}
				continue;
			}
			
			if(hasStarted){
				returnValue.put(k, v);
			}
		}
		return returnValue;
	}
	
	

	@SuppressWarnings("unchecked")
	public static <K1, K2, V1, V2> Map<K2, V2> castMap(Map<K1, V1> map) {
		Map<K2, V2> newMap = new HashMap<K2, V2>();
		for (Map.Entry<K1, V1> entry : map.entrySet()) {
			K2 key = (K2) entry.getKey();
			V2 value = (V2) entry.getValue();
			newMap.put(key, value);
		}
		return newMap;
	}

	public static <K, V> Map<K, V> merge(Map<? extends K, ? extends V> map1,
			Map<? extends K, ? extends V> map2) {
		Map<K, V> returnMap = new HashMap<K, V>();
		returnMap.putAll(map1);
		returnMap.putAll(map2);
		return returnMap;
	}

	public static <K, V> Map<K, V> removeGenericFromMap(
			Map<? extends K, ? extends V> map) {
		Map<K, V> returnMap = new HashMap<K, V>();
		returnMap.putAll(map);
		return returnMap;
	}

	public static <K, MK, MV> Map<K, Map<? extends MK, ? extends MV>> addGenericToDoubleMap(
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> doubleMap) {
		Map<K, Map<? extends MK, ? extends MV>> returnMap = new HashMap<K, Map<? extends MK, ? extends MV>>();
		returnMap.putAll(doubleMap);
		return returnMap;
	}

	public static <K, MK, MV> Map<K, Map<MK, MV>> removeGenericFromDoubleMap(
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> doubleMap) {
		Map<K, Map<MK, MV>> returnMap = new HashMap<K, Map<MK, MV>>();
		for (Map.Entry<? extends K, ? extends Map<? extends MK, ? extends MV>> entry : doubleMap
				.entrySet()) {
			K k = entry.getKey();
			Map<? extends MK, ? extends MV> genericSmallMap = entry.getValue();
			Map<MK, MV> smallMap = removeGenericFromMap(genericSmallMap);
			;
			returnMap.put(k, smallMap);
		}
		return returnMap;
	}

	public static <K, MK, MV> int getDoubleMapSize(
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> doubleMap) {
		if (doubleMap == null) {
			return 0;
		}

		int size = 0;
		for (Map<? extends MK, ? extends MV> valueMap : doubleMap.values()) {
			size += valueMap.size();
		}
		return size;
	}

	public static <K, MK, MV> void mergePostSmallMapToPrevSmallMap(
			Map<K, Map<MK, MV>> prev,
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> post) {

		for (Map.Entry<? extends K, ? extends Map<? extends MK, ? extends MV>> postEntry : post
				.entrySet()) {
			K k = postEntry.getKey();
			Map<? extends MK, ? extends MV> map = postEntry.getValue();
			Map<MK, MV> prevSmallMap = prev.get(k);
			if (prevSmallMap == null) {
				prevSmallMap = new HashMap<MK, MV>();
				prev.put(k, prevSmallMap);
			}
			prevSmallMap.putAll(map);
		}
	}

	public static <K, MK, MV> Map<K, Map<? extends MK, ? extends MV>> mergeSmallMaps(
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> map1,
			Map<? extends K, ? extends Map<? extends MK, ? extends MV>> map2) {

		Map<K, Map<MK, MV>> tmpReturnMap = new HashMap<K, Map<MK, MV>>();
		mergePostSmallMapToPrevSmallMap(tmpReturnMap, map1);
		mergePostSmallMapToPrevSmallMap(tmpReturnMap, map2);

		Map<K, Map<? extends MK, ? extends MV>> returnMap = new HashMap<K, Map<? extends MK, ? extends MV>>();
		returnMap.putAll(tmpReturnMap);

		return returnMap;
	}
	
	public static enum NumericalOperator{
		ADD, SUBTRACT, MULTIPLY, DIVIDE;
		
		public int operate(int a, int b){
			if(this==NumericalOperator.ADD){
				return a+b;
			}else if(this==NumericalOperator.SUBTRACT){
				return a-b;
			}else if(this==NumericalOperator.MULTIPLY){
				return a*b;
			}else if(this==NumericalOperator.DIVIDE){
				return a/b;
			}else{
				throw new RuntimeException("Illegal operator!");
			}
		}
		
		public double operate(Number a, Number b){
			if(this==NumericalOperator.ADD){
				return a.doubleValue()+b.doubleValue();
			}else if(this==NumericalOperator.SUBTRACT){
				return a.doubleValue()-b.doubleValue();
			}else if(this==NumericalOperator.MULTIPLY){
				return a.doubleValue()*b.doubleValue();
			}else if(this==NumericalOperator.DIVIDE){
				return a.doubleValue()/b.doubleValue();
			}else{
				throw new RuntimeException("Illegal operator!");
			}
		}
		
		public double operate(double a, double b){
			if(this==NumericalOperator.ADD){
				return a+b;
			}else if(this==NumericalOperator.SUBTRACT){
				return a-b;
			}else if(this==NumericalOperator.MULTIPLY){
				return a*b;
			}else if(this==NumericalOperator.DIVIDE){
				return a/b;
			}else{
				throw new RuntimeException("Illegal operator!");
			}
		}
	}
	
	public static enum BooleanOperator{
		AND, OR, XOR, XNOR;
		
		public boolean operate(boolean a, boolean b){
			if(this==BooleanOperator.AND){
				return a&&b;
			}else if(this==BooleanOperator.OR){
				return a||b;
			}else if(this==BooleanOperator.XOR){
				return a!=b;
			}else if(this==BooleanOperator.XNOR){
				return a==b;
			}else{
				throw new RuntimeException("Illegal operator!");
			}
		}
	}
	
	public static double[] addEach(double[] a, double[] b) {
		return calculateEach(a,b,NumericalOperator.ADD);
	}
	
	public static double[] calculateEach(double[] v, double a,
			NumericalOperator numericalOperator) {
		double[] returnValue = new double[v.length];
		for(int i=0; i<v.length; i++){
			returnValue[i] = numericalOperator.operate(v[i],a);
		}
		return returnValue;
	}
	
	public static double[] calculateEach(double[] v, Number a,
			NumericalOperator numericalOperator) {
		double[] returnValue = new double[v.length];
		for(int i=0; i<v.length; i++){
			returnValue[i] = numericalOperator.operate(v[i],a);
		}
		return returnValue;
	}

	public static double[] calculateEach(double[] a,
			int[] b, NumericalOperator numericalOperator) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		double[] returnValue = new double[a.length];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = numericalOperator.operate(a[i], b[i]);
		}
		return returnValue;
		
	}

	public static int[] calculateEach(int[] a, int[] b, NumericalOperator numericalOperator) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		int[] returnValue = new int[a.length];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = numericalOperator.operate(a[i], b[i]);
		}
		return returnValue;
	}
	
	public static double[] calculateEach(double[] a, double[] b, NumericalOperator numericalOperator) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		double[] returnValue = new double[a.length];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = numericalOperator.operate(a[i], b[i]);
		}
		return returnValue;
	}
	
	public static double[] calculateEach(double[] a, Number[] b, NumericalOperator numericalOperator) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		double[] returnValue = new double[a.length];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = numericalOperator.operate(a[i], b[i]);
		}
		return returnValue;
	}
	
	public static int[][] calculateEach(int[][] a, int[][] b, NumericalOperator numericalOperator) {
		assertArrayLength(a, b);
		
		int[][] returnValue = new int[a.length][];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = new int[a[i].length];
			for(int j=0; j<a[i].length; j++){
				returnValue[i][j] = numericalOperator.operate(a[i][j], b[i][j]);
			}
		}
		return returnValue;
	}

	public static boolean[] calculateEach(boolean[] a, boolean[] b, BooleanOperator booleanOperator) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		boolean[] returnValue = new boolean[a.length];
		for (int i = 0; i < a.length; i++) {
			returnValue[i] = booleanOperator.operate(a[i], b[i]);
		}
		return returnValue;
	}
	
	public static <T> void assertArrayLength(double[] a, T[] b){
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
	}
	
	public static <T> void assertArrayLength(T[] a, T[] b){
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
	}
	public static <T> void assertArrayLength(T[][] a, T[][] b){
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		for(int i=0; i<a.length; i++){
			T[] aItem = a[i];
			T[] bItem = b[i];
			YeriDebug.ASSERT_compareInteger(aItem.length, bItem.length);
		}
	}
	public static void assertListSize(List<?> a, List<?> b){
		YeriDebug.ASSERT_compareInteger(a.size(), b.size());
		
	}

	public static int[] toIntArray(Collection<? extends Integer> c) {
		Iterator<? extends Integer> iterator = c.iterator();
		int[] returnValue = new int[c.size()];
		for (int i = 0; iterator.hasNext(); i++) {
			returnValue[i] = iterator.next();
		}
		return returnValue;
	}

	public static <T> List<T> toList(T[] tArray) {
		List<T> list = new ArrayList<T>();
		for (T t : tArray) {
			list.add(t);
		}
		return list;
	}

	public static <T> List<T> shuffle(T[] tArray) {
		List<T> list = toList(tArray);
		Collections.shuffle(list);
		return list;
	}

	public static void fillColumn(Object[][] target, String[] data, int column) {
		for (int i = 0; i < data.length; i++) {
			target[i][column] = data[i];
		}
	}
	public static <T extends Comparable<? super T>> int compare(T t1, T t2) {
		if(t1!=null){
			return t1.compareTo(t2);
		}else if(t2!=null){
			return -t2.compareTo(t1);
		}else{
			return 0;
		}
	}
	
	public static int[] getMinMaxIndex(int[] values) {
		int[] indexArray = new int[]{-1, -1};
		int[] valueArray = new int[]{Integer.MAX_VALUE, Integer.MIN_VALUE};
		
		for (int i = 0; i < values.length; i++) {
			int value = values[i];
			if (value < valueArray[0]) {
				indexArray[0] = i;
				valueArray[0] = value;
			}
			if (value > valueArray[1]) {
				indexArray[1] = i;
				valueArray[1] = value;
			}
		}
		return indexArray;
	}
	
	public static int[] getMinMaxValue(int[] values){
		int[] indexArray = getMinMaxIndex(values);
		return new int[]{ values[indexArray[0]], values[indexArray[1]] };
	}

	public static int[] getMinMaxIndex(double[] values) {
		int[] indexArray = new int[]{-1, -1};
		double[] valueArray = new double[]{Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY};
		
		for (int i = 0; i < values.length; i++) {
			double value = values[i];
			if (value < valueArray[0]) {
				indexArray[0] = i;
				valueArray[0] = value;
			}
			if (value > valueArray[1]) {
				indexArray[1] = i;
				valueArray[1] = value;
			}
		}
		return indexArray;
	}
	
	public static <T extends Comparable<? super T>> int[] getMinMaxIndex(List<T> list) {
		
		int[] indexArray = new int[]{-1, -1};
		if(list==null || list.size()<0){ return indexArray; }
		
		T min = list.get(0);
		T max = list.get(0);
		
		for (int i = 1; i < list.size(); i++) {
			T t = list.get(i);
			if(compare(t, min)<0){
				indexArray[0] = i;
				min = t;
			}
			if(compare(t, max)>0){
				indexArray[1] = i;
				max = t;
			}
		}
		return indexArray;
	}
	
	public static <T> int[] getMinMaxIndex(List<? extends T> list, Comparator<? super T> c) {
		
		if(list==null || list.size()<0){ return new int[]{-1, -1}; }
		
		int[] indexArray =  new int[2];
		Arrays.fill(indexArray, 0);
		
		T min = list.get(indexArray[0]);
		T max = list.get(indexArray[1]);
		
		for (int i = 1; i < list.size(); i++) {
			T t = list.get(i);
			
			if(c.compare(min, t)>0){
				indexArray[0] = i;
				min = t;
			}
			if (c.compare(max, t)<0) {
				indexArray[1] = i;
				max = t;
			}
		}
		return indexArray;
	}

	public static <T extends Comparable<? super T>> int[] getMinMaxIndex(T[] tArray) {
			
		int minIndex = -1;
		int maxIndex = -1;
		
		for(int i=0; i<tArray.length; i++){
			T item = tArray[i];
			if(minIndex<0 || item.compareTo(tArray[minIndex])<0){
				minIndex = i;
			}
			if(maxIndex<0 || item.compareTo(tArray[maxIndex])>0){
				maxIndex = i;
			}
		}
		
		return new int[]{ minIndex, maxIndex };
	}


	public static <K, V> void putIfNoKey(Map<K, V> map, K key, V value) {
		if (map.containsKey(key)) {
			return;
		}
		map.put(key, value);
	}

	public static <K, V> Map<K, V> updatePostToPrevious(
			Map<? extends K, ? extends V> previous,
			Map<? extends K, ? extends V> post) {
		return updatePostToPrevious(previous, post, null);
	}

	public static <K, V> Map<K, V> updatePostToPrevious(
			Map<? extends K, ? extends V> previous,
			Map<? extends K, ? extends V> post, Comparator<? super K> c) {
		if (previous == null) {
			return new HashMap<K, V>(post);
		}

		Map<K, V> returnMap = null;
		if (c != null) {
			returnMap = new TreeMap<K, V>(c);
		} else {
			returnMap = new HashMap<K, V>();
		}

		for (K key : post.keySet()) {
			V value;
			if (!previous.containsKey(key)) {
				value = post.get(key);
			} else {
				value = previous.get(key);
			}
			returnMap.put(key, value);
		}
		return returnMap;
	}

	public static <K, V> void putAllIfNoKey(Map<K, V> original,
			Map<? extends K, ? extends V> added) {
		for (Map.Entry<? extends K, ? extends V> entry : added.entrySet()) {
			putIfNoKey(original, entry.getKey(), entry.getValue());
		}
	}

	public static <T> void addAll(Collection<T> collection, T[] array) {
		if (collection == null) {
			return;
		}
		for (T t : array) {
			collection.add(t);
		}
	}
	
	public static void addAll(Collection<Integer> collection, int[] array) {
		if (collection == null) {
			return;
		}
		for (int t : array) {
			collection.add(t);
		}
	}
	
	public static List<Integer> indexListOf(boolean[] array, boolean value) {
		List<Integer> list = new ArrayList<Integer>();
		
		for(int i=0; i<array.length; i++){
			if(array[i]==value){
				list.add(i);
			}
		}
		
		return list;
	}
	
	public static <T> int[] indexArrayOf(T[] array, T[] valueArray) {
		TreeSet<Integer> set = new TreeSet<Integer>();
		
		for(int i=0; i<array.length; i++){
			for(int j=0; j<valueArray.length; j++){
				if(array[i].equals(valueArray[j])){
					set.add(i);
				}
			}
		}
		return CollectionsToolkit.toIntArray(set);
	}

	public static int indexOf(int[] array, int target) {

		for (int i = 0; i < array.length; i++) {
			int t = array[i];
			if (t==target) {
				return i;
			}
		}
		return -1;
	}
	
	public static int[] indexOf(int[][] a, int target) {
		for(int i=0; i<a.length; i++){
			int[] row = a[i];
			int colIndex = CollectionsToolkit.indexOf(row, target);
			if(colIndex>=0){
				return new int[]{i, colIndex};
			}
		}
		return null;
	}
	
	public static <T> int indexOf(T[] array, T target) { return indexOf(array, target, null); }
	public static <T> int indexOf(T[] array, T target, Comparator<? super T> c) {
		if (target == null) {
			return -1;
		}

		for (int i = 0; i < array.length; i++) {
			T t = array[i];
			if(c==null){
				if (t.equals(target)) { return i; }
			}else{
				if (c.compare(t, target)==0) { return i; }
			}
		}
		return -1;
	}

	public static void removeRange(List<? extends Object> list, int fromIndex) {
		removeRange(list, fromIndex, list.size() - 1);
	}

	public static void removeRange(List<? extends Object> list, int fromIndex,
			int toIndex) {
		Iterator<? extends Object> iterator = list.iterator();
		for (int i = 0; iterator.hasNext(); i++) {
			iterator.next();
			if (i >= fromIndex && i <= toIndex) {
				iterator.remove();
			}
		}
	}

	public static void reverse(int[] ds) {
		int tmp;
		int lastIndex = ds.length - 1;
		int reverseI;

		for (int i = 0; i < ds.length / 2; i++) {
			reverseI = lastIndex - 1;
			tmp = ds[i];
			ds[i] = ds[reverseI];
			ds[reverseI] = tmp;
		}
	}
	
	public static void reverse(double[] ds) {
		double tmp;
		int lastIndex = ds.length - 1;
		int reverseI;

		for (int i = 0; i < ds.length / 2; i++) {
			reverseI = lastIndex - 1;
			tmp = ds[i];
			ds[i] = ds[reverseI];
			ds[reverseI] = tmp;
		}
	}
	public static <T> List<T> reverse(List<? extends T> list) {
		List<T> returnList = new ArrayList<T>();
		int listSize = list.size();
		for(int i=listSize-1; i>=0; i--){
			returnList.add(list.get(i));
		}
		return returnList;
	}

	public static <T> List<T> arrayToList(T[] array) {
		List<T> list = new ArrayList<T>();
		for (T t : array) {
			list.add(t);
		}
		return list;
	}

	public static <T> Set<T> arrayToSet(T[] array) {
		Set<T> set = new HashSet<T>();
		for (T t : array) {
			set.add(t);
		}
		return set;
	}

	public static boolean isMapEmpty(Map<?, ?> map) {
		if (map == null || map.size() == 0) {
			return true;
		}
		for (Object o : map.values()) {
			if (o instanceof Map<?, ?>) {
				if (!isMapEmpty((Map<?, ?>) o)) {
					return false;
				}
			} else {
				return false;
			}
		}
		return true;
	}

	public static boolean isListEmpty(List<?> list) {
		return list == null || list.size() == 0;
	}

	public static boolean contains(int[] array, int value) {
		for (int i : array) {
			if (i == value) {
				return true;
			}
		}
		return false;
	}

	public static boolean contains(char[] array, char value) {
		for (int i : array) {
			if (i == value) {
				return true;
			}
		}
		return false;
	}
	
	public static <E> boolean containsByEquals(E[] array, E target) {
		if(array==null){ return false; }
		
		for (E e : array) {
			if (e.equals(target)) {
				return true;
			}
		}
		return false;
	}
	
	public static <E> boolean containsByEquals(Collection<? extends E> c, E target) {
		if(c==null){ return false; }
		
		for (E e : c) {
			if (e.equals(target)) {
				return true;
			}
		}
		return false;
	}
	
	public static <E extends Comparable<? super E>> boolean containsByComparable(
			Collection<? extends E> c, E target) {
		for (E e : c) {
			if (e.compareTo(target) == 0) {
				return true;
			}
		}
		return false;
	}

	public static <E extends Comparable<? super E>> boolean contains(
			Collection<? extends E> c, E target) {
		for (E e : c) {
			if (e.compareTo(target) == 0) {
				return true;
			}
		}
		return false;
	}

	public static <E> boolean contains(Collection<? extends E> c, E target,
			Comparator<? super E> comp) {
		for (E e : c) {
			if(comp!=null){
				if (comp.compare(e, target) == 0) {
					return true;
				}
			}else{
				if (e.equals(target)) {
					return true;
				}
			}
		}
		return false;
	}

	public static <E> Set<E> subtracts(Collection<? extends E> c1,
			Collection<? extends E> c2, Comparator<? super E> comp) {
		HashSet<E> returnValue = new HashSet<E>();
		if (c1 == null) { return null; }
		if (c2 == null) { return new HashSet<E>(c1); }
		for (E o : c1) {
			if (!contains(c2, o, comp)) {
				returnValue.add(o);
			}
		}
		return returnValue;
	}
	
	public static class IndexedElementComparator<E> implements Comparator<IndexedElement<E>>{

		private Comparator<? super E> comparator;
		public IndexedElementComparator(Comparator<? super E> comparator){
			this.comparator = comparator;
		}
		
		@Override
		public int compare(IndexedElement<E> arg0, IndexedElement<E> arg1) {
			return comparator.compare(arg0.getE(), arg1.getE());
		}
	}
	
	
	public static class IndexedElement<E>{
		private E e;
		private int index;
		public IndexedElement(E e, int index){
			this.e = e;
			this.index = index;
		}
		public E getE() {
			return e;
		}
		public int getIndex() {
			return index;
		}
		
		public static <E> Comparator<IndexedElement<E>> createComparator(Comparator<? super E> comparator){
			return new IndexedElementComparator<E>(comparator);
		}
	}
	
	/**
	 * Computationally slow!
	 * Assumptions:
	 *  1. Lists are already sorted. (Though, right now it works without that assumption)
	 *  2. If there is 'null' element in the list, comparator that can handle 'null' is required.
	 */
	public static <E> List<IndexedElement<E>> indexedMerge(List<? extends List<? extends E>> c, Comparator<? super E> comparator) {
		
		int listCount = c.size();
		List<IndexedElement<E>> allList = new ArrayList<IndexedElement<E>>();
		for(int i=0; i<listCount; i++){
			List<? extends E> list = c.get(i);
			for(E e : list){
				IndexedElement<E> indexedElement = new IndexedElement<E>(e, i);
				allList.add(indexedElement);
			}
		}
		Collections.sort(allList, IndexedElement.createComparator(comparator));
		return allList;
	}
	
	/**
	 * Computationally slow!
	 * Assumptions:
	 *  1. Lists are already sorted. (Though, right now it works without that assumption)
	 *  2. If there is 'null' element in the list, comparator that can handle 'null' is required.
	 */
	public static <E extends Comparable<? super E>> List<E> merge(Collection<? extends List<? extends E>> c) {
		
		List<E> allList = new ArrayList<E>();
		for(List<? extends E> list : c){
			allList.addAll(list);
		}
		Collections.sort(allList);
		return allList;
	}
	
	/**
	 * Computationally slow!
	 * Assumptions:
	 *  1. Lists are already sorted. (Though, right now it works without that assumption)
	 *  2. If there is 'null' element in the list, comparator that can handle 'null' is required.
	 */
	public static <E> List<E> merge(Collection<? extends List<? extends E>> c, Comparator<? super E> comparator) {
		
		List<E> allList = new ArrayList<E>();
		for(List<? extends E> list : c){
			allList.addAll(list);
		}
		Collections.sort(allList, comparator);
		return allList;
	}

	public static <E> Set<E> intersection(Collection<? extends E> c1,
			Collection<? extends E> c2, Comparator<? super E> comp) {
		HashSet<E> returnValue = new HashSet<E>();
		if (c1 == null) {
			return new HashSet<E>(c2);
		}else if (c2 == null) {
			return new HashSet<E>(c1);
		}
		for (E o : c1) {
			if(comp!=null){
				if (contains(c2, o, comp)) {
					returnValue.add(o);
				}
			}else{
				if (c2.contains(o)) {
					returnValue.add(o);
				}
			}
		}
		return returnValue;
	}
	
	public static <T> boolean[] contains(T[] tArray, T[] targetArray) {
		boolean[] result = new boolean[targetArray.length];
		for(int i=0; i<targetArray.length; i++){
			result[i] = contains(tArray, targetArray[i]);
		}
		return result;
	}
	public static <T> boolean contains(T[] tArray, T target) {
		return indexOf(tArray, target) >= 0;
	}
	public static <T> boolean containsByComparator(T[] tArray, T target, Comparator<? super T> c) {
		return indexOf(tArray, target, c) >= 0;
	}

	public static <K, V> SortedMap<K, V> getSortedMap(
			Map<? extends K, ? extends V> map, Comparator<? super K> c) {
		SortedMap<K, V> sortedMap = new TreeMap<K, V>(c);
		sortedMap.putAll(map);
		return sortedMap;
	}

	public static void fill(boolean[][] box, boolean b) {
		if (box == null) {
			return;
		}
		for (boolean[] rowArray : box) {
			if (rowArray == null) {
				continue;
			}
			Arrays.fill(rowArray, b);
		}
	}
	
	public static void fill(int[][] box, int value) {
		if (box == null) {
			return;
		}
		for (int[] rowArray : box) {
			if (rowArray == null) {
				continue;
			}
			Arrays.fill(rowArray, value);
		}
	}
	
	public static void fill(int[][][] cube, int value) {
		if (cube == null) {
			return;
		}
		for (int[][] box : cube) {
			fill(box, value);
		}
	}
	
	public static void fill(double[][] array, double value) {
		if (array == null) {
			return;
		}
		for (double[] rowArray : array) {
			if (rowArray == null) {
				continue;
			}
			Arrays.fill(rowArray, value);
		}
	}
	public static void fill(double[][][] array, double value) {
		if (array == null) {
			return;
		}
		for (double[][] eachArray : array) {
			if (eachArray == null) {
				continue;
			}
			fill(eachArray, value);
		}
	}

	public static int[] increEach(int[] values, int change) {
		int[] result = new int[values.length];
		for (int i = 0; i < values.length; i++) {
			int newValue = values[i] + change;
			result[i] = newValue;
		}
		return result;
	}

	public static <T> void fillEach(T[] propertyNames,
			Collection<? extends T> collection) {
		int collectionSize = collection.size();
		int arrayLength = propertyNames.length;
		int limit = Math.min(collectionSize, arrayLength);

		Iterator<? extends T> iterator = collection.iterator();
		int i = 0;
		for (; i < limit; i++) {
			T t = iterator.next();
			propertyNames[i] = t;
		}

		for (; i < arrayLength; i++) {
			propertyNames[i] = null;
		}

	}
	
	public static String toString(int[] floatArray) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[');
		for(int i=0; i<floatArray.length; i++){
			int value = floatArray[i];
			if(i>0){ buffer.append(","); }
			buffer.append(value);
		}
		buffer.append(']');
		return buffer.toString();
	}

	public static String toString(float[] floatArray) {
		return toString(floatArray, null);
	}
	public static String toString(float[] floatArray, DecimalFormat format) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[');
		for(int i=0; i<floatArray.length; i++){
			float f = floatArray[i];
			if(i>0){ buffer.append(","); }
			if(format!=null){
				buffer.append(format.format(f));
			}else{
				buffer.append(f);
			}
		}
		buffer.append(']');
		return buffer.toString();
	}
	
	public static String toString(List<Integer> intList, int displayCountPerLine) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[');
		for(int i=0; i<intList.size(); i++){
			int f = intList.get(i);
			if(i>0){ buffer.append(","); }
			buffer.append(f);
			if( (i+1)%displayCountPerLine==0){
				buffer.append(StringToolkit.newLine());
			}
		}
		buffer.append(']');
		return buffer.toString();
	}
	
	public static String toString(double[] floatArray) {
		return toString(floatArray, null);
	}
	public static String toString(double[] floatArray, DecimalFormat format) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[');
		for(int i=0; i<floatArray.length; i++){
			double d = floatArray[i];
			if(i>0){ buffer.append(","); }
			if(format!=null){
				buffer.append(format.format(d));
			}else{
				buffer.append(d);
			}
		}
		buffer.append(']');
		return buffer.toString();
	}
	
	public static enum ListStringType{
		HORIZONTAL, VERTICAL;

		public String process(String s) {
			if(this==VERTICAL){
				return s+","+StringToolkit.newLine();
			}else{
				return s+",";
			}
		}
	}

	public static <T> String toString(Collection<? extends T> list) {
		return toString(list, ListStringType.HORIZONTAL);
	}
	public static <T> String toString(Collection<? extends T> list, ListStringType listStringType) {
		StringBuffer buffer = new StringBuffer();
		buffer.append('[');
		for(T t : list){
			buffer.append(listStringType.process(t.toString()));
//			buffer.append(t.toString()).append(',');
		}
		buffer.append(']');
		return buffer.toString();
	}

	public static String toString(double[][][] logMArray) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[').append(StringToolkit.newLine());
		for(int i=0; i<logMArray.length; i++){
			double[][] logMi = logMArray[i];
			buffer.append(' ').append('[').append(StringToolkit.newLine());
			for(int j=0; j<logMi.length; j++){
				buffer.append("  ");
				buffer.append(toString(logMi[j]));
				buffer.append(StringToolkit.newLine());
			}
			buffer.append(' ').append(']').append(StringToolkit.newLine());
		}
		buffer.append(']');
		
		return buffer.toString();
	}

	public static String toString(int[][] array2) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append(' ').append('[').append(StringToolkit.newLine());
		for(int j=0; j<array2.length; j++){
			buffer.append("  ");
			buffer.append(toString(array2[j]));
			buffer.append(StringToolkit.newLine());
		}
		buffer.append(' ').append(']').append(StringToolkit.newLine());
		
		return buffer.toString();
	}
	
	public static String toString(int[][][] array3) {
		StringBuffer buffer = new StringBuffer();
		
		buffer.append('[').append(StringToolkit.newLine());
		for(int i=0; i<array3.length; i++){
			int[][] array2 = array3[i];
			buffer.append(toString(array2));
		}
		buffer.append(']');
		
		return buffer.toString();
	}

	public static String toString(double[][] logMi) {
		StringBuffer buffer = new StringBuffer();
		buffer.append('[').append(StringToolkit.newLine());
		for(int i=0; i<logMi.length; i++){
			buffer.append(' ');
			buffer.append(toString(logMi[i]));
			buffer.append(StringToolkit.newLine());
		}
		buffer.append(']');
		return buffer.toString();
	}

	public static boolean isRectangular(boolean[][] box) {
		if(box.length==0){ return true; }
		
		int length = box[0].length;
		for(int i=1; i<box.length; i++){
			boolean[] x = box[i];
			if(x.length!=length){
				return false;
			}
		}
		return true;
	}
	
	public static <T> boolean isRectangular(T[][] a) {
		if(a.length==0){ return true; }
		
		int length = a[0].length;
		for(int i=1; i<a.length; i++){
			T[] x = a[i];
			if(x.length!=length){
				return false;
			}
		}
		return true;
	}
	public static boolean isRectangular(int[][] a) {
		if(a.length==0){ return true; }
		
		int length = a[0].length;
		for(int i=1; i<a.length; i++){
			int[] x = a[i];
			if(x.length!=length){
				return false;
			}
		}
		return true;
	}
	
	public static boolean isRectangular(double[][] a) {
		if(a.length==0){ return true; }
		
		int length = a[0].length;
		for(int i=1; i<a.length; i++){
			double[] x = a[i];
			if(x.length!=length){
				return false;
			}
		}
		return true;
	}

	public static boolean equals(double[] d1, double[] d2) {
		if(d1==null){ return (d2==null); }
		if(d2==null){ return false; }
		
		if(d1.length!=d2.length){ return false; }
		for(int i=0; i<d1.length; i++){
			if(d1[i] != d2[i]){
				return false;
			}
		}
		return true;
	}
	
	public static boolean equals(String[] s1, String[] s2){
		return equals(s1,s2,false);
	}
	public static boolean equals(String[] s1, String[] s2, boolean ignoreCase){
		if(s1==null){ return s2==null; }
		else if(s2==null){ return false; }
		
		if(s1.length!=s2.length){ return false; }
		
		for(int i=0; i<s1.length; i++){
			if(s1[i]==null){
				if(s2[i]!=null){ return false; }
				else{ continue; }
			}
			
			StringComparator c = new StringComparator(ignoreCase);
			if(c.compare(s1[i], s2[i])!=0){ return false; }
			else{ continue; }
		}
		for(String s : s2){
			if(!CollectionsToolkit.containsByComparator(s1, s, new StringComparator(ignoreCase))){
				return false;
			}
		}
		return true;
	}
	
	public static void checkEquals(double[] d1, double[] d2) {
		if(equals(d1,d2)){ return; }
		StringBuffer buffer = new StringBuffer();
		buffer.append("Two arrays not equal!").append(StringToolkit.newLine());
		buffer.append(toString(d1)).append(StringToolkit.newLine());
		buffer.append(toString(d2)).append(StringToolkit.newLine());
		throw new RuntimeException(buffer.toString());
	}

	public static int sum(int[] intArray) {
		int returnValue = 0;
		for(int i : intArray){
			returnValue += i;
		}
		return returnValue;
	}
	
	public static double sum(double[] doubleArray) {
		double returnValue = 0;
		for(double d : doubleArray){
			returnValue += d;
		}
		return returnValue;
	}

	public static String[] subtracts(String[] v, String[] w) {
		List<String> vSubW = new ArrayList<String>();
		for(String word : v){
			if(contains(w, word)){ continue; }
			vSubW.add(word);
		}
		return vSubW.toArray(new String[0]);
	}

	public static int[] sumVertical(
			int[][] tableArray) {
		YeriDebug.ASSERT(CollectionsToolkit.isRectangular(tableArray));
		
		int[] result = new int[tableArray[0].length];
		Arrays.fill(result,0);
		
		for(int j=0; j<tableArray[0].length; j++){
			for(int i=0; i<tableArray.length; i++){
				result[j] += tableArray[i][j];
			}
		}
		
		return result;
	}

	public static <T> String[] toStringArray(List<T> list) {
		String[] returnValue = new String[list.size()];
		for(int i=0; i<list.size(); i++){
			T t = list.get(i);
			returnValue[i] = t.toString();
		}
		return returnValue;
	}

	public static <T> void copy(T[] target, int targetStart, T[] data, int dataStart, int dataEnd) {
		for(int i=0; i<dataEnd-dataStart; i++){
			target[i+targetStart] = data[i+dataStart];
		}
	}

	public static String[] concatenate(String[][] strings) {
		int totalSize = 0;
		for(String[] array : strings){
			totalSize += array.length;
		}
		String[] target = new String[totalSize];
		
		int targetStart = 0;
		for(String[] array : strings){
			copy(target, targetStart, array, 0, array.length);
			targetStart += array.length;
		}
		return target;
	}

	public static boolean hasCommonValue(int[] array1,
			int[] array2) {
		for(int x : array1){
			for(int y : array2){
				if(x==y){ return true; }
			}
		}
		return false;
	}
	
	public static int[] toIntArray(List<Integer> intList){
		int[] returnValue = new int[intList.size()];
		for(int i=0; i<intList.size(); i++){
			int value = intList.get(i);
			returnValue[i] = value;
		}
		return returnValue;
	}
	
	public static int[] getCommonValueArray(int[] array1,
			int[] array2) {
		List<Integer> commonValue = new ArrayList<Integer>();
		for(int x : array1){
			for(int y : array2){
				if(x==y){ commonValue.add(x); }
			}
		}
		return toIntArray(commonValue);
	}

	public static <T> Collection<T> flattenCollection(
			Collection<? extends Collection<? extends T>> values) {
		List<T> returnValue = new ArrayList<T>();
		for(Collection<? extends T> c : values){
			returnValue.addAll(c);
		}
		return returnValue;
	}

	public static <T> void copy(T[][] source, T[][] target) {
		assertArrayLength(source, target);
		for(int i=0; i<source.length; i++){
			for(int j=0; j<source[i].length; j++){
				target[i][j] = source[i][j];
			}
		}
	}
	public static <K,V> Map<K, V> createHashMap(K[] kArray, V[] vArray) {
		YeriDebug.ASSERT(kArray.length==vArray.length);
		
		Map<K,V> map = new HashMap<K,V>();
		for(int i=0; i<kArray.length; i++){
			map.put(kArray[i], vArray[i]);
		}
		return map;
	}
	
	public static CharSequence toString(boolean[][] box, char[] cArray) {
		StringBuilder builder = new StringBuilder();
		for(int i=0; i<box.length; i++){
			for(int j=0; j<box[i].length; j++){
				builder.append(cArray[box[i][j]?0:1]);
			}
			builder.append(StringToolkit.newLine());
		}
		return builder;
	}
	public static <T> T[] getObjectsAtIndices(T[] tArray,
			int[] indices) {
		if(indices==null){ return tArray; }
		T[] returnArray = Arrays.copyOf(tArray, indices.length);
		for(int i=0; i<indices.length; i++){
			int index = indices[i];
			returnArray[i] = tArray[index];
		}
		return returnArray;
	}
	
	public static <K,V> void removeMapEntryNotInList(Map<? extends K,V> map, Collection<? extends K> kCollection) {
		if(kCollection==null){ return; }
		List<K> kList = new ArrayList<K>(map.keySet());
		for(K k : kList){
			if(!kCollection.contains(k)){
				map.remove(k);
			}
		}
	}
	
	public static <K,V> void fillEmptyMapEntryWithDefaultValue(Map<K, V> map, Collection<? extends K> kCollection, V defaultV) {
		if(kCollection==null){ return; }
		
		for(K k : kCollection){
			if(!map.containsKey(k)){
				map.put(k, defaultV);
			}
		}
	}
	
	public static class ScoreWrapper<T> extends Number{
		private static final long serialVersionUID = 1L;
		
		public String toString(){
			StringBuilder builder = new StringBuilder();
			builder.append(key.toString()).append("=").append(score);
			return builder.toString();
		}
		private T key;
		private double score;
		
		public T getKey() {
			return key;
		}

		public double getScore() {
			return score;
		}

		public ScoreWrapper(T key, double score){
			this.key = key;
			this.score = score;
		}
		
		@SuppressWarnings("unchecked")
		public boolean equals(Object o){
			if(!(o instanceof ScoreWrapper)){ return false; }
			
			ScoreWrapper sw = (ScoreWrapper)o;
			return this.key.equals(sw.key);
		}

		@Override
		public double doubleValue() {
			return score;
		}

		@Override
		public float floatValue() {
			return (float)score;
		}

		@Override
		public int intValue() {
			return (int)score;
		}

		@Override
		public long longValue() {
			return (long)score;
		}

		
		@SuppressWarnings("unchecked")
		public static <T> int getComparedValue(ScoreWrapper<T> s1, ScoreWrapper<T> s2, double diff){
			if(diff!=0){ return diff<0?-1:1; }
			
			T s1Key = s1.getKey();
			if(s1Key instanceof Comparable){
				Comparable c = (Comparable)s1Key;
				return c.compareTo(s2.getKey());
			}
			else{ return -1; } 
		}
		
		@SuppressWarnings("unchecked")
		public static <T> Comparator<ScoreWrapper<T>> getComparator(){
			return new Comparator<ScoreWrapper<T>>(){
				public int compare(ScoreWrapper<T> s1, ScoreWrapper<T> s2){
					double diff = s1.score - s2.score;
					return getComparedValue(s1,s2,diff);
				}
			};
		}
		
		public static <T> Comparator<ScoreWrapper<T>> getAbsoluteComparator(){
			return new Comparator<ScoreWrapper<T>>(){
				public int compare(ScoreWrapper<T> s1, ScoreWrapper<T> s2){
					double diff = Math.abs(s1.score) - Math.abs(s2.score);
					return getComparedValue(s1,s2,diff);
				}
			};
		}
		
		@SuppressWarnings("unchecked")
		public static <T> Comparator<ScoreWrapper<T>> getReverseComparator(){
			return new Comparator<ScoreWrapper<T>>(){
				public int compare(ScoreWrapper<T> s1, ScoreWrapper<T> s2){
					double diff = -(s1.score - s2.score);
					return getComparedValue(s1,s2,diff);
				}
			};
		}
		
		@SuppressWarnings("unchecked")
		public static <T> Comparator<ScoreWrapper<T>> getReverseAbsoluteComparator(){
			return new Comparator<ScoreWrapper<T>>(){
				public int compare(ScoreWrapper<T> s1, ScoreWrapper<T> s2){
					double diff = -(Math.abs(s1.score) - Math.abs(s2.score));
					return getComparedValue(s1,s2,diff);
				}
			};
		}
	}
	public static Comparator<Number> getAscendNumberComparator(){
		return new Comparator<Number>(){
			@Override
			public int compare(
					Number o1,
					Number o2) {
				double diff = o1.doubleValue() - o2.doubleValue(); 
				return (diff<0)?-1:(diff==0?0:1);
			}
		};
	}
	public static Comparator<Number> getDescendNumberComparator(){
		return createReverseComparator(getAscendNumberComparator());
	}
	
	public static void main(String[] args){
//		System.out.println(0==Double.NaN);
		test01();
	}
	
	public static CharSequence toString(Map<String, ?> map) {
		StringBuilder builder = new StringBuilder();
		for(String key : map.keySet()){
			builder.append(key);
			builder.append(' ');
			builder.append(map.get(key));
			builder.append(StringToolkit.newLine());
		}
		return builder;
	}
	
	private static class EntryValueComparator<K> implements Comparator<Map.Entry<K, ? extends Number>>{

		@Override
		public int compare(Entry<K, ? extends Number> o1,
				Entry<K, ? extends Number> o2) {
			return StatisticsToolkit.getSign(o1.getValue().doubleValue()-o2.getValue().doubleValue());
		}
	}
	public static <K> List<Map.Entry<K, ? extends Number>> getSortedEntryList(Map<K,? extends Number> map, boolean asc){
		List<Map.Entry<K, ? extends Number>> entryList = new ArrayList<Map.Entry<K, ? extends Number>>(map.entrySet());
		EntryValueComparator<K> c = new EntryValueComparator<K>();
		if(asc){ Collections.sort(entryList, c); }
		else{ Collections.sort(entryList, createReverseComparator(c) ); }
		return entryList;
	}
	
	private static class ComparatorByMapValue<K> implements Comparator<K>{
		private Map<K,? extends Number> map;
		public ComparatorByMapValue(Map<K, ? extends Number> map){
			this.map = map;
		}

		@Override
		public int compare(K k1, K k2) {
			Number n1 = map.get(k1);
			double d1 = n1==null?0:n1.doubleValue();
			Number n2 = map.get(k2);
			double d2 = n2==null?0:n2.doubleValue();
			return StatisticsToolkit.getSign(d1-d2);
		}
	}
	public static <K> List<K> getSortedEntryList(Collection<? extends K> keySet, Map<K, ? extends Number> map, boolean asc) {
		List<K> keyList = new ArrayList<K>(keySet);
		
		ComparatorByMapValue<K> c = new ComparatorByMapValue<K>(map);
		if(asc){ Collections.sort(keyList, c); }
		else{ Collections.sort(keyList, createReverseComparator(c) ); }
		return keyList;
		
	}
	public static List<Character> toList(char[] charArray) {
		List<Character> characterList = new ArrayList<Character>();
		for(char c : charArray){
			characterList.add(c);
		}
		return characterList;
	}
	public static CharSequence toString(Object[] a, String delim) {
		StringBuilder builder = new StringBuilder();
		for(int i=0; i<a.length; i++){
			if(i>0){ builder.append(delim); }
			builder.append(a[i].toString());
		}
		return builder;
	}
	public static int indexOf(String[] valueArray,
			String value, boolean ignoreCase) {
		if(value==null){ return -1; }
		
		for(int i=0; i<valueArray.length; i++){
			String currentValue = valueArray[i];
			if(ignoreCase){
				if(value.equalsIgnoreCase(currentValue)){
					return i;
				}
			}else{
				if(value.equals(currentValue)){
					return i;
				}
			}
		}
		return -1;
	}
	public static int indexOfEndsWith(String[] endsArray,
			String value, boolean ignoreCase) {
		if(value==null){ return -1; }
		
		for(int i=0; i<endsArray.length; i++){
			String end = endsArray[i];
			if(ignoreCase){
				if( StringToolkit.endsWithIgnoreCase(value,end) ){
					return i;
				}
			}else{
				if(value.endsWith(end)){
					return i;
				}
			}
		}
		return -1;
	}
	public static void copy(double[] target, int targetStart, double[] data, int dataStart, int dataEnd) {
		for(int i=0; i<dataEnd-dataStart; i++){
			target[i+targetStart] = data[i+dataStart];
		}
	}

	public static double[] concatenate(double[][] ds) {
		int totalLength =0;
		for(double[] d : ds){
			totalLength += d.length;
		}
		double[] target = new double[totalLength];
		
		int targetStart = 0;
		for(double[] d: ds){
			copy(target, targetStart, d, 0, d.length);
			targetStart += d.length;
		}
		return target;
	}
	
	public static double[] getRank(double[] values) {
		return getRank(values, true);
	}
	public static double[] getRank(double[] values, boolean smallToLarge) {
		double[] copy = Arrays.copyOf(values, values.length);
		Arrays.sort(copy);
		
		if(!smallToLarge){
			reverse(copy);
		}
		
		double[] rankArray = new double[values.length];
		Arrays.fill(rankArray, Double.NaN);
		
		double prevValue = Double.NaN;
		for(int i=0; i<copy.length; i++){
			int rank = i+1;
			double value = copy[i];
			if(prevValue==value){ continue; }
			prevValue = value;
			
			int[] indexArray = CollectionsToolkit.indexArrayOf(values, value);
			double realRank = rank+ ((double)(indexArray.length-1))/2;
			for(int index : indexArray){
				rankArray[index] = realRank;
			}
		}
		return rankArray;
	}
	
	public static int[] getOrderIndex(int[] values) {
		return getOrderIndex(values, true);
	}
	public static int[] getOrderIndex(int[] values, boolean smallToLarge) {
		int[] copy = Arrays.copyOf(values, values.length);
		Arrays.sort(copy);
		
		if(!smallToLarge){
			reverse(copy);
		}
		
		int[] orderIndexArray = new int[values.length];
		Arrays.fill(orderIndexArray, -1);
		
		Integer prevValue = null;
		for(int i=0; i<copy.length; ){
			int value = copy[i];
			if(prevValue!=null && prevValue==value){ continue; }
			prevValue = value;
			
			int[] indexArray = CollectionsToolkit.indexArrayOf(values, value);
			
			for(int index : indexArray){
				orderIndexArray[index] = i;
				i++;
			}
		}
		return orderIndexArray;
	}
	
	public static int[] getOrderIndex(double[] values) {
		return getOrderIndex(values, true);
	}
	public static int[] getOrderIndex(double[] values, boolean smallToLarge) {
		double[] copy = Arrays.copyOf(values, values.length);
		Arrays.sort(copy);
		
		if(!smallToLarge){
			reverse(copy);
		}
		
		int[] orderIndexArray = new int[values.length];
		Arrays.fill(orderIndexArray, -1);
		
		double prevValue = Double.NaN;
		for(int i=0; i<copy.length; ){
			double value = copy[i];
			if(prevValue==value){ continue; }
			prevValue = value;
			
			int[] indexArray = CollectionsToolkit.indexArrayOf(values, value);
			
			for(int index : indexArray){
				orderIndexArray[index] = i;
				i++;
			}
		}
		return orderIndexArray;
	}
	
	private static int[] indexArrayOf(int[] values, int value) {
		List<Integer> indexList = new ArrayList<Integer>();
		for(int i=0; i<values.length; i++){
			int thisValue = values[i];
			if(thisValue==value){
				indexList.add(i);
			}
		}
		return toIntArray(indexList);
	}
	private static int[] indexArrayOf(double[] values, double value) {
		List<Integer> indexList = new ArrayList<Integer>();
		for(int i=0; i<values.length; i++){
			double thisValue = values[i];
			if(thisValue==value){
				indexList.add(i);
			}
		}
		return toIntArray(indexList);
	}
	
	protected static void test01() {
		double[] values = new double[]{1,2,3,2,2,4,5,7,4};
		int[] orderArray = getOrderIndex(values);
		System.out.println(StringToolkit.toString(orderArray));
	}
	public static int[] getIndexArrayOfValueArray(int[] arrayOfValueSet,
			int[] valueListArray) {
		int[] indexArray = new int[valueListArray.length];
		
		for(int i=0; i<valueListArray.length; i++){
			indexArray[i] = Arrays.binarySearch(arrayOfValueSet, valueListArray[i])+1;
		}
		return indexArray;
	}
	public static <T> int[] getIndexArrayOfValueArray(T[] arrayOfValueSet, T[] valueListArray) {
		return getIndexArrayOfValueArray(arrayOfValueSet, valueListArray, null);
	}
	private static <T> int[] getIndexArrayOfValueArray(T[] arrayOfValueSet, T[] valueListArray, Comparator<? super T> c) {
		if(valueListArray==null){ return new int[0]; }
		
		int[] indexArray = new int[valueListArray.length];
		
		for(int i=0; i<valueListArray.length; i++){
			T t = valueListArray[i];
			indexArray[i] = CollectionsToolkit.indexOf(arrayOfValueSet, t, c);
		}
		return indexArray;
	}
	
	private static int adjustNegativeIndexToSegmentStartIndex(int index){
		if(index>=0){ return index; }
		else{
			return -(index+1)-1;
		}
	}
	public static boolean isInSameSegment(int[] segmentationEndIndexArray, int[] positionArray){
		
		int index=0;
		for(int i=0; i<positionArray.length; i++){
			int currentIndex = adjustNegativeIndexToSegmentStartIndex(Arrays.binarySearch(segmentationEndIndexArray, positionArray[i]));
			if(i!=0){
				if(currentIndex!=index){
					return false;
				}
			}
			index = currentIndex;
		}
		return true;
	}
	public static <T> List<T> copyOf(List<T> valueList, int start, int n) {
		List<T> returnList = new ArrayList<T>();
		for(int i=0; i<n; i++){
			returnList.add(valueList.get(start+i));
		}
		return returnList;
	}
	public static int[] getReverseOrderIndex(int[] orderIndex) {
		int[] reverseOrderIndex = new int[orderIndex.length];
		
		for(int i=0; i<orderIndex.length; i++){
			reverseOrderIndex[i] = orderIndex.length - orderIndex[i] - 1;
		}
		return reverseOrderIndex;
	}
	
	public static <T> boolean equals(T[] a, T[] b){
		if(a==null){
			return b==null;
		}else if(b==null){
			return false;
		}
		if(a.length!=b.length){ return false; }
		for(int i=0; i<a.length; i++){
			if(!a[i].equals(b[i])){ return false; }
		}
		return true;
	}
	public static int[] union(int[] a, int[] b) {
		TreeSet<Integer> set = new TreeSet<Integer>();
		CollectionsToolkit.addAll(set, a);
		CollectionsToolkit.addAll(set, b);
		return CollectionsToolkit.toIntArray(set);
	}
	
	public static <T> List<T> filterOutItems(T[] array, Collection<? extends T> removalCollection) {
		Set<T> removalSet = new TreeSet<T>(removalCollection);
		List<T> returnList = new ArrayList<T>();
		
		for(T t : array){
			if(removalSet.contains(t)){ continue; }
			
			returnList.add(t);
		}
		return returnList;
	}
	
	public static <T> List<T> getItemList(List<T> itemList, int[] indexArray){
		List<T> returnList = new ArrayList<T>();
		for(int index : indexArray){
			returnList.add(itemList.get(index));
		}
		return returnList;
	}
	public static <T> boolean[] equalsEach(T[] a, T[] b) {
		YeriDebug.ASSERT_compareInteger(a.length, b.length);
		
		boolean[] result = new boolean[a.length];
		for(int i=0; i<a.length; i++){
			result[i] = ((a[i]==b[i]) || a[i].equals(b[i])); 
		}
		return result;
	}
	public static <T> List<T> getItemListOfValue(T[] tArray, boolean[] valueArray, boolean value) {
		List<T> itemList = new ArrayList<T>();
		
		for(int i=0; i<valueArray.length; i++){
			if(valueArray[i]!=value){ continue; }
			itemList.add(tArray[i]);
		}
		return itemList;
	}
	public static <T,U> List<T> getItemListOfValue(T[] tArray, U[] valueArray, U value) {
		List<T> itemList = new ArrayList<T>();
		
		for(int i=0; i<valueArray.length; i++){
			if( !valueArray[i].equals(value) ){ continue; }
			itemList.add(tArray[i]);
		}
		return itemList;
	}
	public static double[] getColumnValueArray(double[][] doubleArray, int index) {
		double[] returnValue = new double[doubleArray.length];
		for(int i=0; i<returnValue.length; i++){
			returnValue[i] = doubleArray[i][index];
		}
		return returnValue;
	}
	public static <T> String[] getClassNameArray(Collection<T> list) {
		String[] classNameArray = new String[list.size()];
		
		Iterator<T> iterator = list.iterator();
		for(int i=0; iterator.hasNext(); i++){
			T t = iterator.next();
			classNameArray[i] = t.getClass().getName();
		}
		return classNameArray;
	}
	public static boolean calculateUp(boolean[] contains, BooleanOperator op) {
		YeriDebug.ASSERT(contains!=null && contains.length>0);
		
		boolean b = contains[0];
		for(int i=1; i<contains.length; i++){
			b = op.operate(b,contains[i]);
		}
		return b;
	}
	public static void removeAll(List<Thread> notifyThreadList, Thread[] array) {
		for(Thread t : array){
			notifyThreadList.remove(t);
		}
	}
	public static boolean[] toBooleanArray(Collection<? extends Boolean> c) {
		boolean[] returnValue = new boolean[c.size()];
		
		Iterator<? extends Boolean> iterator = c.iterator();
		for(int i=0; i<c.size(); i++){
			returnValue[i] = iterator.next().booleanValue();
		}
		return returnValue;
	}
	public static <T> T[] add(T[] array, T t) {
		return add(array, t, array.length);
	}
	public static <T> T[] add(T[] array, T t, int index) {
		T[] returnValue = Arrays.copyOfRange(array, array.length, 2*array.length+1);
		
		for(int i=0; i<index; i++){ returnValue[i] = array[i]; }
		returnValue[index] = t;
		for(int i=index; i<array.length; i++){ returnValue[i+1] = array[i]; }
		return returnValue;
	}
	public static <T> int[][] getTabularCount(T[] trueTArray, T[] predictedTArray, T[] itemArray) {
		YeriDebug.ASSERT_compareInteger(trueTArray.length, predictedTArray.length);
		
		int itemCount = itemArray.length;
		int[][] returnArray = new int[itemCount][itemCount];
		
		for(int i=0; i<trueTArray.length; i++){
			int trueIndex = CollectionsToolkit.indexOf(itemArray, trueTArray[i]);
			int predictedIndex = CollectionsToolkit.indexOf(itemArray, predictedTArray[i]);
			returnArray[trueIndex][predictedIndex]++;
		}
		return returnArray;
	}
	public static <T> T[] removeDuplicates(T[] array) {
		Set<T> set = new TreeSet<T>();
		T[] returnArray = Arrays.copyOf(array, 0);
		addAll(set, array);
		return set.toArray(returnArray);
	}
	
	public static <T> int[] getElementIndexArrayOfThisClass(T[] array, T t) {
		if(t==null || array==null){ return null; }
		
		List<Integer> indexList = new ArrayList<Integer>();
		
		for(int i=0; i<array.length; i++){
			T thisT = array[i];
			if(RTTIToolkit.isPreviousObjectInstanceOfRearObjectClass(thisT, t)){ indexList.add(i); }
		}
		
		
		return toIntArray(indexList);
	}
	
	public static <T> T[] getElementArrayOfThisClass(T[] array, T t) {
		int[] indexArray = getElementIndexArrayOfThisClass(array,t);
		if(indexArray==null){ return null; }
		
		T[] returnArray = Arrays.copyOf(array, indexArray.length);
		
		for(int i=0; i<indexArray.length; i++){
			int index = indexArray[i];
			returnArray[i] = array[index];
		}
		return returnArray;
	}
	
	public static <T> T[] updateObjectsOfSameClass(T[] original, T[] updates){
		
		T[] returnArray = Arrays.copyOf(original, original.length);
		if(updates==null || updates.length==0){ return returnArray; }
		
		for(int i=0; i<updates.length; i++){
			
			int[] indexArray = getElementIndexArrayOfThisClass(returnArray, updates[i]);
			if(indexArray==null || indexArray.length==0){ continue; }
			
			for(int j=0; j<indexArray.length; j++){
				returnArray[indexArray[j]] = updates[i];
			}
		}
		return returnArray;
	}
	public static Integer[] toWrapperArray(int[] array) {
		Integer[] returnArray = new Integer[array.length];
		for(int i=0; i<array.length; i++){
			returnArray[i] = array[i];
		}
		return returnArray;
	}
	
	public static class MapUsingComparator<K,V> implements Comparator<K>{

		private Map<K,V> map;
		private Comparator<? super V> c;
		private MapUsingComparator(Map<K,V> map, Comparator<? super V> c){
			this.map = map;
			this.c = c;
		}
		
		public static <K,V> MapUsingComparator<K,V> create(Map<K,V> map, Comparator<? super V> c){
			return new MapUsingComparator<K,V>(map, c);
		}
		
		@Override
		public int compare(K k1, K k2) {
			V v1 = map.get(k1);
			V v2 = map.get(k2);
			return c.compare(v1, v2);
		}
		
	}
	
	public static <K,V> List<K> sortKeyByValue(Map<K, V> map, Comparator<? super V> c) {
		MapUsingComparator<K,V> comparator = MapUsingComparator.create(map, c);
		List<K> list = new ArrayList<K>(map.keySet());
		
		Collections.sort(list, comparator);
		return list;
	}
}
