package com.yerihyo.yeritools.collections;

import java.io.Serializable;
import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;

/**
 * This class implements a heap.
 */
public class MinHeap<T> extends AbstractCollection<T>  implements Cloneable, Serializable{
	
	private static final long serialVersionUID = 1L;
	
	private Comparator<? super T> comparator = null;
	
	public void setComparator(Comparator<? super T> comparator) {
		this.comparator = comparator;
	}

	@SuppressWarnings("unchecked")
	protected int compare(T t1, T t2){
		if(comparator==null){
			if(t1 instanceof Comparable){
				return ((Comparable<T>)t1).compareTo(t2);
			}else{
				throw new RuntimeException("Uncomparable types in MinHeap");
			}
		}else{
			return comparator.compare(t1, t2);
		}
	}

	/**
	 * Constructs an empty heap.
	 * 
	 */
	public MinHeap() {
		elements = new ArrayList<T>();
		elements.add(null);
	}

	public MinHeap(Comparator<? super T> c) {
		this();
		this.setComparator(c);
	}

	/**
	 * Adds a new element to this heap.
	 * 
	 * @param newElement
	 *            the element to add
	 */
	public boolean add(T t) {
		// Add a new leaf
		elements.add(null);
		int index = elements.size() - 1;

		// Demote parents that are larger than the new element
		while (index > 1 && compare(getParent(index),t) > 0) {
			elements.set(index, getParent(index));
			index = getParentIndex(index);
		}

		// Store the new element into the vacant slot
		elements.set(index, t);
		return true;
	}
	
	public boolean addAll(Collection<? extends T> tList){
		boolean b = true;
		for(T t : tList){
			b = b && this.add(t);
		}
		return b;
	}

	/**
	 * Gets the minimum element stored in this heap.
	 * 
	 * @return the minimum element
	 */
	public T peek() {
		return elements.get(1);
	}

	/**
	 * Removes the minimum element from this heap.
	 * 
	 * @return the minimum element
	 */
	public T getMinimum() {
		T minimum = elements.get(1);

		// Remove last element
		int lastIndex = elements.size() - 1;
		T last = elements.remove(lastIndex);

		if (lastIndex > 1) {
			elements.set(1, last);
			fixHeap();
		}

		return minimum;
	}

	/**
	 * Turns the tree back into a heap, provided only the root node violates the
	 * heap condition.
	 */
	private void fixHeap() {
		T root = elements.get(1);

		int lastIndex = elements.size() - 1;
		// Promote children of removed root while they are larger than last

		int index = 1;
		boolean more = true;
		while (more) {
			int childIndex = getLeftChildIndex(index);
			if (childIndex <= lastIndex) {
				// Get smaller child

				// Get left child first
				T child = getLeftChild(index);

				// Use right child instead if it is smaller
				if (getRightChildIndex(index) <= lastIndex
						&& compare(getRightChild(index),child) < 0) {
					childIndex = getRightChildIndex(index);
					child = getRightChild(index);
				}

				// Check if larger child is smaller than root
				if (compare(child,root) < 0) {
					// Promote child
					elements.set(index, child);
					index = childIndex;
				} else {
					// Root is smaller than both children
					more = false;
				}
			} else {
				// No children
				more = false;
			}
		}

		// Store root element in vacant slot
		elements.set(index, root);
	}

	/**
	 * Returns the number of elements in this heap.
	 */
	public int size() {
		return elements.size() - 1;
	}

	/**
	 * Returns the index of the left child.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the index of the left child of the given node
	 */
	private static int getLeftChildIndex(int index) {
		return 2 * index;
	}

	/**
	 * Returns the index of the right child.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the index of the right child of the given node
	 */
	private static int getRightChildIndex(int index) {
		return 2 * index + 1;
	}

	/**
	 * Returns the index of the parent.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the index of the parent of the given node
	 */
	private static int getParentIndex(int index) {
		return index / 2;
	}

	/**
	 * Returns the value of the left child.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the value of the left child of the given node
	 */
	private T getLeftChild(int index) {
		return elements.get(2 * index);
	}

	/**
	 * Returns the value of the right child.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the value of the right child of the given node
	 */
	private T getRightChild(int index) {
		return elements.get(2 * index + 1);
	}

	/**
	 * Returns the value of the parent.
	 * 
	 * @param index
	 *            the index of a node in this heap
	 * @return the value of the parent of the given node
	 */
	private T getParent(int index) {
		return elements.get(index / 2);
	}

	private ArrayList<T> elements;
	
	@Override
	public Iterator<T> iterator() {
		Iterator<T> iterator = elements.iterator();
		iterator.next();
		return iterator;
	}
}
