package com.yerihyo.yeritools.text;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.yerihyo.yeritools.collections.CollectionsToolkit;
import com.yerihyo.yeritools.math.StatisticsToolkit;
import com.yerihyo.yeritools.text.LevenshteinDistance.LevensteinDistanceAnalysis.Operation;

public class LevenshteinDistance {
	public static class LevensteinDistanceAnalysis{
		public static enum Operation{
			INSERT, DELETE, IDENTICAL, SUBSTITUTE;
			
			public String toString(){
				return this.name();
			}
			public static short[] getAllInsertCountArray(int insertCount){
				short[] countArray = new short[Operation.values().length];
				countArray[0] = (short)insertCount;
				return countArray;
			}
		}
		
		private short[][] distance;
		private Operation[][] operations;
		public LevensteinDistanceAnalysis(short[][] distance, Operation[][] operations){
			this.distance = distance;
			this.operations = operations;
		}
		
		public short[][] getDistance() {
			return distance;
		}
		public Operation[][] getOperations() {
			return operations;
		}
		
		public String toString(){
			StringBuilder builder = new StringBuilder();
			boolean isFirst = true;
			for(Operation operation: this.getViterbiOperationArray()){
				if(isFirst){
					isFirst = false;
				}else{
					builder.append("-");
				}
				builder.append(operation.name());
			}
			return builder.toString();
		}
		private Operation[] viterbiOperationArray = null;
		public Operation[] getViterbiOperationArray(){
			if(viterbiOperationArray==null){
				List<Operation> operationList = new ArrayList<Operation>();
				for(int i=distance.length-1, j=distance[0].length-1; i>0 || j>0; ){
					Operation op = operations[i][j];
					operationList.add(op);
					if(op==Operation.INSERT){ j--; }
					else if(op==Operation.DELETE){ i--; }
					else{ i--; j--; }
				}
				Collections.reverse(operationList);
				viterbiOperationArray = operationList.toArray(new Operation[0]);
			}
			return viterbiOperationArray;
		}
		
		public int getViterbiOperationCount(Operation operation){
			int count = 0;
			for(Operation curOperation : this.getViterbiOperationArray()){
				if(curOperation==operation){
					count++;
				}
			}
			return count;
		}
		
		public short[] getViterbiOperationCountArray(){
			Operation[] operationArray = Operation.values();
			short[] countArray = new short[operationArray.length];
			
			for(Operation operation : this.getViterbiOperationArray()){
				int index = CollectionsToolkit.indexOf(operationArray, operation);
				countArray[index]++;
			}
			return countArray;
		}
		
		
		public <T> List<T> getRemainingObjects(T[] s, T[] t){
			List<T> returnList = new ArrayList<T>();
			Operation[] operationArray = this.getViterbiOperationArray();
			for(int i=0, j=0, o=0; o<operationArray.length; o++){
				Operation op = operationArray[o];
				if(op==Operation.INSERT){
					returnList.add(t[j++]);
				}else if(op==Operation.DELETE){
					i++;
				}else if(op==Operation.IDENTICAL){
					returnList.add(s[i++]);
					j++;
				}else if(op==Operation.SUBSTITUTE){
					i++;
					returnList.add(t[j++]);
				}
			}
			return returnList;
		}
		
		public <T> List<T> getRemainingObjects(List<T> s, List<T> t){
			List<T> returnList = new ArrayList<T>();
			Operation[] operationArray = this.getViterbiOperationArray();
			for(int i=0, j=0, o=0; o<operationArray.length; o++){
				Operation op = operationArray[o];
				if(op==Operation.INSERT){
					returnList.add(t.get(j++));
				}else if(op==Operation.DELETE){
					i++;
				}else if(op==Operation.IDENTICAL){
					returnList.add(s.get(i++));
					j++;
				}else if(op==Operation.SUBSTITUTE){
					i++;
					returnList.add(t.get(j++));
				}
			}
			return returnList;
		}
		
		public <T> List<T> getNewObjects(T[] t){
			List<T> returnList = new ArrayList<T>();
			Operation[] operationArray = this.getViterbiOperationArray();
			for(int j=0, o=0; o<operationArray.length; o++){
				Operation op = operationArray[o];
				if(op==Operation.INSERT){
					returnList.add(t[j++]);
				}else if(op==Operation.DELETE){
				}else if(op==Operation.IDENTICAL){
					j++;
				}else if(op==Operation.SUBSTITUTE){
					returnList.add(t[j++]);
				}
			}
			return returnList;
		}
		
		public <T> List<T> getRemovedObjects(T[] s){
			List<T> returnList = new ArrayList<T>();
			Operation[] operationArray = this.getViterbiOperationArray();
			for(int i=0, j=0, o=0; o<operationArray.length; o++){
				Operation op = operationArray[o];
				if(op==Operation.DELETE){
					returnList.add(s[i++]);
				}else if(op==Operation.IDENTICAL){
					i++; j++;
				}else if(op==Operation.SUBSTITUTE){
					returnList.add(s[i++]);
				}
			}
			return returnList;
		}
		
		public <T> List<T> getRemovedObjects(List<T> s){
			List<T> returnList = new ArrayList<T>();
			Operation[] operationArray = this.getViterbiOperationArray();
			for(int i=0, j=0, o=0; o<operationArray.length; o++){
				Operation op = operationArray[o];
				if(op==Operation.DELETE){
					returnList.add(s.get(i++));
				}else if(op==Operation.IDENTICAL){
					i++; j++;
				}else if(op==Operation.SUBSTITUTE){
					returnList.add(s.get(i++));
				}
			}
			return returnList;
		}
	}
	
	public static void main(String[] args){
		test01();
	}
	
	private static class TestObject{
		private int i;
		private String s;
		
		public TestObject(int i, String s){
			this.i =i;
			this.s = s;
		}
		
		public static Comparator<TestObject> getComparator(){
			return new Comparator<TestObject>(){
				private Comparator<String> c = new StringToolkit.StringComparator(true);

				public int compare(TestObject o1, TestObject o2) {
					return c.compare(o1.s, o2.s);
				}
				
			};
		}
		
		public String toString(){
			return "("+i+","+s+")";
		}
	}
	
	protected static void test01() {
		TestObject[] a = new TestObject[]{
				new TestObject(1,"a"),
				new TestObject(1,"b"),
				new TestObject(1,"b"),
				new TestObject(1,"a"),
		};
		TestObject[] b = new TestObject[]{
				new TestObject(2,"b"),
				new TestObject(2,"a"),
				new TestObject(2,"c"),
				new TestObject(2,"b"),
		};
		LevensteinDistanceAnalysis analysis = getLevensteinDistance(a,b,(short)1,(short)1,(short)1, TestObject.getComparator());
		System.out.println( CollectionsToolkit.toString(analysis.getViterbiOperationArray(),"-") );
		System.out.println( CollectionsToolkit.toString(analysis.getRemovedObjects(a).toArray(new TestObject[0]),"-") );
		System.out.println( CollectionsToolkit.toString(analysis.getRemainingObjects(a,b).toArray(new TestObject[0]),"-") );
	}

	public static int getLevensteinDistance(String s, String t) {
		return getLevensteinDistance(s,t, 1,1,1);
	}
	public static int getLevensteinDistance(String s, String t, int insCost, int delCost, int subCost) {
		
		int sLength= s.length();
		int tLength= t.length();
		int[][] d = new int[sLength + 1][tLength + 1];
		d[0][0] = 0;

		for (int i = 0; i <= sLength; i++) {
			for (int j = 0; j <= tLength; j++) {
				if(i==0 && j==0){ continue; }
				
				int deleteValue = (i==0)?Integer.MAX_VALUE:(d[i - 1][j] + delCost);
				int insertValue = (j==0)?Integer.MAX_VALUE:(d[i][j - 1] + insCost);
				int identicalValue =  (i!=0&&j!=0&&(s.charAt(i-1)==t.charAt(j-1)))?d[i - 1][j-1]:Integer.MAX_VALUE;
				int substituteValue = (i!=0&&j!=0&&(s.charAt(i-1)!=t.charAt(j-1)))?d[i - 1][j-1]+subCost:Integer.MAX_VALUE; 
				int[] valueArray = new int[]{
						insertValue, deleteValue, identicalValue, substituteValue,
				};
				
				int[] minmaxIndex = StatisticsToolkit.getMinMaxIndex(valueArray);
				int minIndex = minmaxIndex[0];
				
				d[i][j] = valueArray[minIndex];
			}

		}

		// Step 7
		return d[sLength][tLength];
	}
	
	public static <T> LevensteinDistanceAnalysis getLevensteinDistance(T[] s, T[] t, short insCost, short delCost, short subCost, Comparator<? super T> c) {
		
		int sLength= s.length;
		int tLength= t.length;
//		YeriDebug.getYeriDebug().traceln("Levenstein distance of "+(sLength+1)+"*"+(tLength+1));
		
		Operation[][] op = new Operation[sLength + 1][tLength + 1];
		short[][] d = new short[sLength + 1][tLength + 1];
		d[0][0] = 0;

		for (int i = 0; i <= sLength; i++) {
			for (int j = 0; j <= tLength; j++) {
				if(i==0 && j==0){ continue; }
				
				short deleteValue = (i==0)?Short.MAX_VALUE:(short)(d[i - 1][j] + delCost);
				short insertValue = (j==0)?Short.MAX_VALUE:(short)(d[i][j - 1] + insCost);
				short identicalValue =  (i!=0&&j!=0&&(c.compare(s[i-1],t[j-1])==0))?d[i - 1][j-1]:Short.MAX_VALUE;
				short substituteValue = (i!=0&&j!=0&&(c.compare(s[i-1],t[j-1])!=0))?(short)(d[i - 1][j-1]+subCost):Short.MAX_VALUE; 
				short[] valueArray = new short[]{
						insertValue, deleteValue, identicalValue, substituteValue,
				};
				
				int minIndex = StatisticsToolkit.getMinMaxIndex(valueArray)[0];
				
				d[i][j] = valueArray[minIndex];
				op[i][j] = Operation.values()[minIndex]; 
			}

		}
		
		return new LevensteinDistanceAnalysis(d,op);
	}
	public static <T> LevensteinDistanceAnalysis getLevensteinDistance(List<T> s, List<T> t, short insCost, short delCost, short subCost, Comparator<? super T> c) {
		
		int sLength= s.size();
		int tLength= t.size();
		Operation[][] op = new Operation[sLength + 1][tLength + 1];
		short[][] d = new short[sLength + 1][tLength + 1];
		d[0][0] = 0;

		for (int i = 0; i <= sLength; i++) {
			for (int j = 0; j <= tLength; j++) {
				if(i==0 && j==0){ continue; }
				
				short deleteValue = (i==0)?Short.MAX_VALUE:(short)(d[i - 1][j] + delCost);
				short insertValue = (j==0)?Short.MAX_VALUE:(short)(d[i][j - 1] + insCost);
				short identicalValue =  (i!=0&&j!=0&&(c.compare(s.get(i-1),t.get(j-1))==0))?0:Short.MAX_VALUE;
				short substituteValue = (i!=0&&j!=0&&(c.compare(s.get(i-1),t.get(j-1))!=0))?(short)(d[i - 1][j-1] + subCost):Short.MAX_VALUE; 
				short[] valueArray = new short[]{
						insertValue, deleteValue, identicalValue, substituteValue,
				};
				
				int minIndex = StatisticsToolkit.getMinMaxIndex(valueArray)[0];
				
				d[i][j] = valueArray[minIndex];
				op[i][j] = Operation.values()[minIndex]; 
			}

		}
		
		return new LevensteinDistanceAnalysis(d,op);
	}
	
	

}
